/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "wparameters.hh"
#include "matlab_fcs.hh"
#include <iostream>
#include <cstdlib>
#include <cstdio>

using namespace wpipe;
using namespace std;

//======================================  Construct parameter list from file.
wparameters::wparameters(const std::string& parFile, int debugLevel) {

   //-----------------------------------  Get default parameter list file.
   string parameterFile = parFile;
   if (parameterFile.empty()) parameterFile = "./parameters.txt";

  // validate command line arguments
  if (!exist(parameterFile, "file")) {
    error(string("could not find parameter file: ") + parameterFile);
  }

  /////////////////////////////////////////////////////////////////////////////
  //                            initialize parameters                        //
  /////////////////////////////////////////////////////////////////////////////

  // initialize parameters
  sampleFrequency = 0;
  maximumMismatch = 0;
  falseEventRate = 0;
  blockDuration = 0;
  conditionDuration = 0;
  errorOnStateError = false;
  highPassCutoff = -1.0;
  lowPassCutoff = -1.0;
  whiteningDuration = 0;
  transientFactor = 4.0;
  doubleWhiten = 1;
  extraBlockOverlap = 0;
  outlierFactor = 2.0;
  maximumSignificants = 100000;
  maximumTriggers = 1000;
  durationInflation = 1.0;
  bandwidthInflation = 1.0;
  randomSeed = 0.0;

  applyClustering  = false;
  clusterRadius    = 0.0;
  clusterDensity   = 0.0;
  clusterSingles   = 0;
  clusterThreshold = 0.0;
  writeClusters    = 0;

  coincidenceNumber  = 0;
  maximumCoincidents = 999999999;
  
  applyVeto            = 1;
  falseVetoRate        = 0;
  uncertaintyFactor    = 0;
  correlationFactor    = 0;
  vetoDurationFactor   = 0.5;
  vetoBandwidthFactor  = 0.5;
  maximumConsistents   = 1000;


  bayesian = false;
#ifndef NO_FOLLOWUP
  bayesian = true;
  maxFollowTriggers = 5;
  writeSkymap = false;
  gzipSkymap = false;
  snrRatioCut = 50.0;

  xCoherentCheck = false;
  maxEnt = false;

  combinedStat = false;
  //combinedParameters = [];
#endif

  /////////////////////////////////////////////////////////////////////////////
  //                               read parameters                           //
  /////////////////////////////////////////////////////////////////////////////

  // open parameter file for reading
  FILE* parameterFileID = fopen(parameterFile.c_str(), "r");

  // begin loop over parameter file
  while (!feof(parameterFileID)) {
    char temp[1024];
    char* line = fgets(temp, sizeof(temp), parameterFileID);
    if (!line) continue;
    
    // read one line from parameter file
    string parameterLine = line;

    // remove any comments
    string::size_type commentIndex = parameterLine.find("#");
    string::size_type altIndex     = parameterLine.find("%");
    if (altIndex < commentIndex) commentIndex = altIndex;
    altIndex                       = parameterLine.find("//");
    if (altIndex < commentIndex) commentIndex = altIndex;
    if (commentIndex != string::npos) parameterLine.erase(commentIndex);

    // remove leading and trailing blanks
    parameterLine = deblank(parameterLine);

    // if (empty line, skip to the next line
    if (parameterLine.empty()) continue;

    // locate field separator
    string::size_type colonIndex = parameterLine.find(":");

    // if (field separator not located, report syntax error
    if (colonIndex == string::npos) {
      error(string("syntax error processing parameter file: ")+parameterLine);
    }

    // parse parameter line
    string parameterName  = parameterLine.substr(0, colonIndex);
    string parameterValue = parameterLine.substr(colonIndex + 1);
    parameterName = deblank(parameterName);
    parameterValue = deblank(parameterValue);

    // report parameter settings
    wlog(debugLevel, 1, parameterName + ": " + parameterValue);

    // assign parameters based on name
    // required parameters
    if (parameterName == "channelNames") {
      channelNames = eval_str(parameterValue);
    } else if (parameterName ==  "frameTypes") {
      frameTypes = eval_str(parameterValue);
    }

    // optional parameters
    else if (parameterName ==  "analysisMode") {
      analysisMode = unquote(parameterValue);
    } else if (parameterName ==  "sampleFrequency") {
      sampleFrequency = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "qRange") {
      qRange = eval(parameterValue);
    } else if (parameterName ==  "frequencyRange") {
      frequencyRange = eval(parameterValue);
    } else if (parameterName ==  "maximumMismatch") {
      maximumMismatch = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "falseEventRate") {
      falseEventRate = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "blockDuration") {
      blockDuration = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "conditionDuration") {
      conditionDuration = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "dataFactors") {
      dataFactors = eval(parameterValue);
    } else if (parameterName ==  "timeShifts") {
      timeShifts = eval(parameterValue);
    } else if (parameterName ==  "stateNames") {
      stateNames = eval_str(parameterValue);
    } else if (parameterName ==  "stateTypes") {
      stateTypes = eval_str(parameterValue);
    } else if (parameterName ==  "stateMasks") {
      stateMasks = eval(parameterValue);
    } else if (parameterName ==  "errorOnStateError") {
      errorOnStateError = strtol(parameterValue.c_str(), 0, 0);
    } else if (parameterName ==  "injectionNames") {
      injectionNames = eval_str(parameterValue);
    } else if (parameterName ==  "injectionTypes") {
      injectionTypes = eval_str(parameterValue);
    } else if (parameterName ==  "injectionFactors") {
      injectionFactors = eval(parameterValue);
    } else if (parameterName ==  "injectionTimeShifts") {
      injectionTimeShifts = eval(parameterValue);
    } else if (parameterName ==  "highPassCutoff") {
      highPassCutoff = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "lowPassCutoff") {
      lowPassCutoff = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "whiteningDuration") {
      whiteningDuration = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "transientFactor") {
      transientFactor = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "doubleWhiten") {
      doubleWhiten = strtol(parameterValue.c_str(), 0, 0);
    } else if (parameterName ==  "extraBlockOverlap") {
      extraBlockOverlap = strtod(parameterValue.c_str(), 0);
    } 
    else if (parameterName ==  "outlierFactor") {
      outlierFactor = strtod(parameterValue.c_str(), 0);
    }
    //----------------------------------  Input maximum significants
    //  Note: large ints are read in as doubles to allow exp notation (e.g. 1e5)
    else if (parameterName ==  "maximumSignificants") {
       maximumSignificants = int(strtod(parameterValue.c_str(), 0));
    }
    else if (parameterName ==  "maximumTriggers") {
       maximumTriggers = int(strtod(parameterValue.c_str(), 0));
    } 
    //-----------------------------------
    else if (parameterName ==  "durationInflation") {
      durationInflation = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "bandwidthInflation") {
      bandwidthInflation = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "triggerFields") {
      triggerFields = eval_str(parameterValue);
    } else if (parameterName ==  "triggerFormat") {
      triggerFormat = unquote(parameterValue);
    } else if (parameterName ==  "randomSeed") {
      randomSeed = strtod(parameterValue.c_str(), 0);
    }

    // clustering parameters
    else if (parameterName ==  "applyClustering") {
      applyClustering = strtol(parameterValue.c_str(), 0, 0);
    } else if (parameterName ==  "clusterMethod") {
      clusterMethod = unquote(parameterValue);
    } else if (parameterName ==  "clusterRadius") {
      clusterRadius = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "clusterDensity") {
      clusterDensity = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "clusterSingles") {
      clusterSingles = strtol(parameterValue.c_str(), 0, 0);
    } else if (parameterName ==  "clusterLinkage") {
      clusterLinkage = unquote(parameterValue);
    } else if (parameterName ==  "clusterCriterion") {
      clusterCriterion = unquote(parameterValue);
    } else if (parameterName ==  "clusterThreshold") {
      clusterThreshold = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "distanceMetric") {
      distanceMetric = unquote(parameterValue);
    } else if (parameterName ==  "writeClusters") {
      writeClusters = strtol(parameterValue.c_str(), 0, 0);
    }

    // "independent" mode parameters
    else if (parameterName ==  "coincidenceNumber") {
      coincidenceNumber = strtol(parameterValue.c_str(), 0, 0);
    } else if (parameterName ==  "maximumCoincidents") {
      maximumCoincidents = strtol(parameterValue.c_str(), 0, 0);
    }

#ifndef NO_COHERENT
    // "coherent" mode parameters
    else if (parameterName ==  "skyPosition") {
      skyPosition = eval(parameterValue);
    } else if (parameterName ==  "skyCoordinateSystem") {
      skyCoordinateSystem = unquote(parameterValue);
    }
#endif

    // coherent mode parameters used in function calls.
    else if (parameterName ==  "applyVeto") {
      applyVeto = strtol(parameterValue.c_str(), 0, 0);
    } else if (parameterName ==  "falseVetoRate") {
      falseVetoRate = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "uncertaintyFactor") {
      uncertaintyFactor = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "correlationFactor") {
      correlationFactor = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "vetoDurationFactor") {
      vetoDurationFactor = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "vetoBandwidthFactor") {
      vetoBandwidthFactor = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "maximumConsistents") {
      maximumConsistents = strtol(parameterValue.c_str(), 0, 0);
    }

#ifndef NO_FOLLOWUP
    // "bayesian" mode parameters
    else if (parameterName ==  "bayesian") {
      bayesian = strtol(parameterValue.c_str(), 0, 0);
    } else if (parameterName ==  "maxFollowTriggers") {
      maxFollowTriggers = strtol(parameterValue.c_str(), 0, 0);
    } else if (parameterName ==  "snrRatioCut") {
      snrRatioCut = strtod(parameterValue.c_str(), 0);
    } else if (parameterName ==  "writeSkymap") {
      writeSkymap = strtol(parameterValue.c_str(), 0, 0);
    } else if (parameterName ==  "gzipSkymap") {
      gzipSkymap = strtol(parameterValue.c_str(), 0, 0);
    } else if (parameterName ==  "prcInjectionFile") {
      prcInjectionFile = unquote(parameterValue);
    }

    // xCoherentCheck
    else if (parameterName ==  "xCoherentCheck") {
      xCoherentCheck = strtol(parameterValue.c_str(), 0, 0);
    }

    // MaxEnt waveform recovery
    else if (parameterName ==  "maxEnt") {
      maxEnt = strtol(parameterValue.c_str(), 0, 0);
    }

    // Combined statistic
    else if (parameterName ==  "combinedStat") {
      combinedStat = strtol(parameterValue.c_str(), 0, 0);
    } else if (parameterName ==  "combinedParameters") {
      combinedParameters = eval(parameterValue);
    }
#endif

    // handle unknown parameters
    else if (debugLevel >= 1) {
      cout <<  "unknown parameter " << parameterName << ".  skipping." << endl;
    }  // end assign parameters based on name

  } // end loop over parameter file entries

  // close parameter file
  fclose(parameterFileID);

  /////////////////////////////////////////////////////////////////////////////
  //                        check for required parameters                    //
  /////////////////////////////////////////////////////////////////////////////

  // test for unspecified parameters
  if (channelNames.empty()) {
    error("channelNames not specified");
  }
  if (frameTypes.empty()) {
    error("frameTypes not specified");
  }

  /////////////////////////////////////////////////////////////////////////////
  //                              set unset defaults                         //
  /////////////////////////////////////////////////////////////////////////////

  // number of sites
  sites.clear();
  for (size_t i=0; i<numberOfChannels(); i++) {
    string isite = channelNames[i].substr(0, 1);
    if (sites.find(isite) == string::npos) sites += isite;
  }

  // default values
  if (analysisMode.empty()) analysisMode="independent";

  if (sampleFrequency == 0) sampleFrequency = 4096;

  if (qRange.empty()) {
    double qDefault[2] = {sqrt(11), 100.0};
    qRange = dble_vect(qDefault, qDefault+2);
  }

  if (frequencyRange.empty()) {
    double fDefault[2] = {48.0, 1.0/0.0};
    frequencyRange = dble_vect(fDefault, fDefault+2);
  }

  if (maximumMismatch == 0) maximumMismatch = 0.2;

  if (falseEventRate == 0) falseEventRate = 1;

  if (blockDuration == 0)  blockDuration = 64;

  if (conditionDuration == 0) conditionDuration = blockDuration;

  if (dataFactors.empty()) {
    dataFactors = dble_vect(numberOfChannels(), 1.0);
  }

  if (timeShifts.empty()) {
    timeShifts =  dble_vect(numberOfChannels(), 0.0);
  }

  if (injectionNames.empty()) {
    injectionNames.resize(numberOfChannels());
  }

  if (injectionTypes.empty()) {
     injectionTypes.resize(numberOfChannels(), "NONE");
  }

  if (injectionFactors.empty()) {
    injectionFactors.resize(numberOfChannels(), 0.0);
  }

  if (injectionTimeShifts.empty()) {
    injectionTimeShifts = dble_vect(numberOfChannels(), 0.0);
  }

  if (triggerFormat.empty()) {
    triggerFormat = "txt";
  }

  //if (randomSeed == 0) {
  //  randomSeed = sum(1e6 * clock);
  //}

  // defaults for clustering
  if (clusterMethod.empty()) {
    clusterMethod = "density";
  }
  if (clusterMethod == "density") {
    if (clusterRadius  == 0) clusterRadius  = 4.0;
    if (clusterDensity == 0) clusterDensity = 3.0;
    if (clusterSingles == 0) clusterSingles = 1;
  }
  else if (clusterMethod == "hierarchical") {
    if (clusterLinkage.empty())   clusterLinkage   = "single";
    if (clusterCriterion.empty()) clusterCriterion = "distance";
    if (clusterThreshold == 0.0)  clusterThreshold = 4.0;
  }

  if (distanceMetric.empty()) {
    distanceMetric = "integratedMismatch";
  }

#ifndef NO_COHERENT
  // defaults for coherent analysis
  if (skyCoordinateSystem.empty()) {
    skyCoordinateSystem = "equatorial";
  }

  if (combinedParameters.empty()) {
    double defCPar[2] = {1.1, 1.0};
    combinedParameters = dble_vect(defCPar, defCPar+2);
  }
#endif
  
  // default trigger fields
  if (triggerFields.empty()) {
    const char* defFields[5] = {"time", "frequency", "duration", 
				"bandwidth", "normalizedEnergy"};

    triggerFields = str_vect(defFields, defFields+5);

    if (analysisMode == "coherent") {
      triggerFields.push_back("incoherentEnergy");
    }

    if (applyClustering && !writeClusters) {
      triggerFields.push_back("clusterSize");
      triggerFields.push_back("clusterNormalizedEnergy");
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  //                             validate parameters                         //
  /////////////////////////////////////////////////////////////////////////////

  size_t nChan = numberOfChannels();
  if (analysisMode == "coherent") {
    if (nChan < 2) {
      error("coherent analysis mode requires at least two channels");
    }
  } else if (analysisMode != "independent") {
    error(string("unknown analysis mode ") + analysisMode);
  }

  // validate number of frame types
  if (frameTypes.size() != nChan) {
    error("number of frameTypes and channelNames are inconsistent");
  }

  // validate number of data factors
  if (dataFactors.size() != nChan && !dataFactors.empty()) {
    error("number of dataFactors and channelNames are inconsistent");
  }

  // validate number of time shifts
  if (timeShifts.size() != nChan && !timeShifts.empty()) {
    error("number of timeShifts and channelNames are inconsistent");
  }
  for (size_t i=0; i<nChan; ++i) {
    timeShifts[i] = long(timeShifts[i]*sampleFrequency + 0.5) / sampleFrequency;
  }

  // validate number of state names
  if (stateNames.size() != nChan && !stateNames.empty()) {
    error("number of stateNames and channelNames are inconsistent");
  }

  // validate number of state types
  if (stateTypes.size() != nChan && !stateTypes.empty()) {
    error("number of stateTypes and channelNames are inconsistent");
  }

  // validate number of state masks
  if (stateMasks.size() != nChan && !stateMasks.empty()) {
    error("number of stateMasks and channelNames are inconsistent");
  }

  // validate number of injection names
  if (injectionNames.size() != nChan && !injectionNames.empty()) {
    error("number of injectionNames and channelNames are inconsistent");
  }

  // validate number of injection types
  if (injectionTypes.size() != nChan && !injectionTypes.empty()) {
    error("number of injectionTypes and channelNames are inconsistent");
  }

  // validate number of injection factors
  if (injectionFactors.size() != nChan && !injectionFactors.empty()) {
    error("number of injectionFactors and channelNames are inconsistent");
  }

  // validate number of injection time shifts
  if (injectionTimeShifts.size() != nChan && !injectionTimeShifts.empty()) {
    error("number of injectionTimeShifts and channelNames are inconsistent");
  }

  // validate block duration
  if (blockDuration != double(int(blockDuration))) {
    error("block duration must be an integer");
  }

#ifndef NO_TARGETED
  // validate sky coordinate system and position
  if (skyCoordinateSystem != "equatorial" &&
      skyCoordinateSystem != "geocentric" &&
      skyCoordinateSystem != "galactic") {
    error("unknown sky coordinate system");
  }
#endif

#ifndef NO_FOLLOWUP
  // check prc injection file location
  if (bayesian) {
    if (nChan < 2) {
      error("bayesian followup requires at least two channels");
    }
    if (coincidenceNumber < 2) {
      error("bayesian followup requires coincidenceNumber > 1");
    }
    if (!prcInjectionFile.empty() && !exist(prcInjectionFile,"file") ) {
      error(string("PRC injection file ")+prcInjectionFile+" not found.");
    }
  }
#endif

  /////////////////////////////////////////////////////////////////////////////
  //                          return to calling function                     //
  /////////////////////////////////////////////////////////////////////////////
}

//======================================  Display parameter values
static std::ostream&
disp_bvalue(std::ostream& out, const std::string& name, bool b) {
   string tab;
   tab.resize(31 - name.size(), ' ');
   out << name << ":" << tab << b << endl;
   return out;
}

static std::ostream&
disp_darray(std::ostream& out, const std::string& name, const dble_vect& d) {
   string tab;
   tab.resize(31 - name.size(), ' ');
   out << name << ":" << tab << "[";
   for (size_t i=0; i<d.size(); i++) {
      if (i) out << " ";
      out << d[i];
   }
   out << "]" << endl;
   return out;
}

static std::ostream&
disp_dvalue(std::ostream& out, const std::string& name, double d) {
   string tab;
   tab.resize(31 - name.size(), ' ');
   out << name << ":" << tab << d << endl;
   return out;
}

static std::ostream&
disp_sarray(std::ostream& out, const std::string& name, const str_vect& s) {
   string tab;
   tab.resize(31 - name.size(), ' ');
   out << name << ":" << tab << "{";
   for (size_t i=0; i<s.size(); i++) {
      if (i) out << " ";
      out << "'" << s[i] << "'";
   }
   out << "}" << endl;
   return out;
}

static std::ostream&
disp_svalue(std::ostream& out, const std::string& name, const std::string& s) {
   string tab;
   tab.resize(31 - name.size(), ' ');
   out << name << ":" << tab << "'" << s << "'" << endl;
   return out;
}

void
wparameters::display(std::ostream& out) const {
   //-----------------------------------  Input data.
   disp_sarray(out, "channelNames",         channelNames);
   disp_sarray(out, "frameTypes",           frameTypes);
   
   //-----------------------------------  Control parameters.
   disp_svalue(out, "analysisMode",        analysisMode);
   disp_dvalue(out, "sampleFrequency",     sampleFrequency);
   disp_darray(out, "qRange",              qRange);
   disp_darray(out, "frequencyRange",      frequencyRange);
   disp_dvalue(out, "maximumMismatch",     maximumMismatch);
   disp_dvalue(out, "falseEventRate",      falseEventRate);
   disp_dvalue(out, "blockDuration",       blockDuration);
   disp_dvalue(out, "conditionDuration",   conditionDuration);
   disp_darray(out, "dataFactors",         dataFactors);
   disp_darray(out, "timeShifts",          timeShifts);

   //-----------------------------------  State information parameters
   disp_sarray(out, "stateNames",          stateNames);
   disp_sarray(out, "stateTypes",          stateTypes);
   disp_darray(out, "stateMasks",          stateMasks);
   disp_bvalue(out, "errorOnStateError",   errorOnStateError);
   disp_sarray(out, "injectionNames",      injectionNames);
   disp_sarray(out, "injectionTypes",      injectionTypes);
   disp_darray(out, "injectionFactors",    injectionFactors);
   disp_darray(out, "injectionTimeShifts", injectionTimeShifts);
   disp_dvalue(out, "highPassCutoff",      highPassCutoff);
   disp_dvalue(out, "lowPassCutoff",       lowPassCutoff);
   disp_dvalue(out, "whiteningDuration",   whiteningDuration);
   disp_dvalue(out, "transientFactor",     transientFactor);
   disp_bvalue(out, "doubleWhiten",        doubleWhiten);
   disp_dvalue(out, "extraBlockOverlap",   extraBlockOverlap);
   disp_dvalue(out, "outlierFactor",       outlierFactor);
   disp_dvalue(out, "falseVetoRate",        falseVetoRate);
   disp_dvalue(out, "uncertaintyFactor",    uncertaintyFactor);
   disp_dvalue(out, "correlationFactor",    correlationFactor);
   disp_dvalue(out, "maximumSignificants", maximumSignificants);
   disp_dvalue(out, "maximumTriggers",     maximumTriggers);
   disp_dvalue(out, "durationInflation",   durationInflation);
   disp_dvalue(out, "bandwidthInflation",  bandwidthInflation);
   disp_sarray(out, "triggerFields",       triggerFields);
   disp_svalue(out, "triggerFormat",       triggerFormat);
   disp_dvalue(out, "randomSeed",          randomSeed);
   disp_dvalue(out, "numberOfChannels",    numberOfChannels());
   disp_dvalue(out, "numberOfSites",       numberOfSites());
   disp_dvalue(out, "injectionChannels",   injectionChannels());


   //-----------------------------------  Clustering parameters.
   disp_dvalue(out, "applyClustering",     applyClustering);
   disp_svalue(out, "clusterMethod",       clusterMethod);
   disp_dvalue(out, "clusterRadius",       clusterRadius);
   disp_dvalue(out, "clusterDensity",      clusterDensity);
   disp_dvalue(out, "clusterSingles",      clusterSingles);
   disp_svalue(out, "clusterLinkage",      clusterLinkage);
   disp_svalue(out, "clusterCriterion",    clusterCriterion);
   disp_dvalue(out, "clusterThreshold",    clusterThreshold);
   disp_svalue(out, "distanceMetric",      distanceMetric);
   disp_dvalue(out, "writeClusters",       writeClusters);

   disp_dvalue(out, "coincidenceNumber",   coincidenceNumber);
   disp_dvalue(out, "maximumCoincidents",  maximumCoincidents);

#ifndef NO_TARGETED
   //-----------------------------------  Optional fields for targeted searches:
   disp_darray(out, "skyPosition",         skyPosition);
   disp_svalue(out, "skyCoordinateSystem", skyCoordinateSystem);
#endif // !defined (NO_TARGETED)

   //-----------------------------------  Fields for 'coherent' analysis
   disp_bvalue(out, "applyVeto",            applyVeto);
   disp_dvalue(out, "vetoDurationFactor",   vetoDurationFactor);
   disp_dvalue(out, "vetoBandwidthFactor",  vetoBandwidthFactor);
   disp_dvalue(out, "maximumConsistents",   maximumConsistents);

   // 'bayesian' position reconstruction:
   disp_bvalue(out, "bayesian",             bayesian);
#ifndef NO_FOLLOWUP
   //-----------------------------------  Parameters for followup analyses.
   disp_dvalue(out, "maxFollowTriggers",    maxFollowTriggers);
   disp_dvalue(out, "snrRatioCut",          snrRatioCut);
   disp_bvalue(out, "writeSkymap",          writeSkymap);
   disp_bvalue(out, "gzipSkymap",           gzipSkymap);
   disp_svalue(out, "prcInjectionFile",     prcInjectionFile);

   // xCoherentCheck null energy statistic:
   disp_bvalue(out, "xCoherentCheck",       xCoherentCheck);

   // MaxEnt waveform reconstruction:
   disp_bvalue(out, "maxEnt",               maxEnt);

   // Combined event statistic:
   disp_bvalue(out, "combinedStat",         combinedStat);
   disp_darray(out, "combinedParameters",   combinedParameters);
#endif // !defined (NO_FOLLOWUP)

   //-----------------------------------  Extra first character of each channel
   disp_svalue(out, "sites",                sites);
}

//======================================  Count injections.
bool
wparameters::inject(size_t inx) const {
    return injectionNames[inx] != "NONE" &&
           injectionTypes[inx] != "NONE" &&
           injectionFactors[inx] != 0.0;
}

//======================================  Count injections.
size_t
wparameters::injectionChannels(void) const {
  size_t nInj = 0;
  size_t N = numberOfChannels();
  for (size_t i=0; i<N; i++) {
    if (inject(i)) nInj++;
  }
  return nInj;
}
