////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//                                  hMon.cc                                   //
//                                                                            //
//      Generation of time-domain calibrated data and analysis segments       //
//                                                                            //
//                            Shourov K. Chatterji                            //
//                                                                            //
//                                2008-Aug-07                                 //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

// $Id: hMon.cc 6444 2011-06-07 21:33:13Z keith.thorne@LIGO.ORG $

// include header files
#include "hMon.hh"
#include "Dacc.hh"
#include "ParseLine.hh"
#include "Segment.hh"
#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <stdlib.h>
#include <cstdio>
#include <cmath>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

// processor identifier for triggers
#define PIDCVSHDR "$Header: https://redoubt.ligo-wa.caltech.edu/svn/gds/trunk/Monitors/hMon/hMon.cc 6444 2011-06-07 21:33:13Z keith.thorne@LIGO.ORG $"
#define PIDTITLE  "generator of real-time time-domain calibrated data"
#include "ProcIdent.hh"

// generate main routine
EXECDAT(hMon)

// macro for log output
#define LOG(OSTREAM, LEVEL) if (verboseLevel >= LEVEL) OSTREAM << logStamp(LEVEL)

/******************************************************************************/
/*                              hMon constructor                              */
/******************************************************************************/

hMon::hMon(int argc, const char *argv[])
  : DatEnv(argc, argv), detectorStates(getDacc())
{

  // test for exceptions
  try {

    // read command line arguments
    readArguments(argc, argv);

    // Set default parameter values
    defaultParameters();

    // Read parameter values from configuration file
    readConfiguration();

    // Determine derived parameter values
    derivedParameters();

    // Test for invalid parameter values
    validateParameters();

    // Initialize the data accessor
    initializeData();

    // Initialize calibated spectrum
    initializeSpectrum();

    // Initialize dmtviewer server
    initializeServer();

    // Initialize minute and second trends
    initializeTrends();

    // Initialize filter coefficients
    initializeFilters();

    // Initialize detector state definitions
    initializeStates();

    // Initialize segments
    initializeSegments();

    // return to calling function
    return;

  }

  // handle non-fatal exceptions
  catch (hMonWarning& warning) {
    return;
  }

  // handle fatal exceptions
  catch (hMonError& error) {
    finish();
    return;
  }

}

/******************************************************************************/
/*                              hMon desctructor                              */
/******************************************************************************/

hMon::~hMon()
{

  // test if analysis has started
  if (lastReadTime != Time(0)) {

    // finalize segment lists
    finalizeSegments();

    // Publish segment lists
    publishSegments();

  }

  // report status
  LOG(std::cout, 1) << "done" << std::endl;

  // return to calling function
  return;

}

/******************************************************************************/
/*                                ProcessData                                 */
/******************************************************************************/

void
hMon::ProcessData(void) {

  // test for exceptions
  try {

    // Read next unit stride of data
    readData();

    // Test for excessive latency
    testLatency();

    // Update segment lists
    updateSegments();

    // Compute optical gain factors
    computeFactors();

    // Test if frame stride complete
    if (isFrameStride()) {

      // compute time-domain calibrated data
      calibrateData();

      // compute calibrated spectra
      computeSpectrum();

      // compute range to fiducial source
      computeSensitivity();

      // Serve data for dmtviewer
      updateServer();

      // Publish calibrated data frames
      publishFrames();

      // Publish segment lists
      publishSegments();

      // Update minute and second trends
      updateTrends();

    // end test for complete frame stride
    }

    // return to calling function
    return;

  }

  // handle non-fatal exceptions
  catch (hMonWarning& warning) {
    return;
  }

  // handle fatal exceptions
  catch (hMonError& error) {
    finish();
    return;
  }

}

/******************************************************************************/
/*                                 Attention                                  */
/******************************************************************************/

void
hMon::Attention(void) {

  // report status
  LOG(std::cout, 2) << "handling attention interrupt" << std::endl;

  // handle server interrupts
  //--- Removed added SIGPOLL test Keith Thorne June, 2011
  // on Linux, Solaris SIGPOLL same as SIGIO, Mac OS X only has SIGIO 
  if (testAttnSignal(SIGIO) == true)  {
    LOG(std::cout, 3) << "handling dmtviewer reqeust" << std::endl;
    MonServer::Attention();
  }

  // handle SIGUSR1 interrupts
  else if (testAttnSignal(SIGUSR1) == true) {
    LOG(std::cout, 3) << "handling user reqeust" << std::endl;
    LOG(std::cerr, 0) << "ERROR: SIGUSR1 handling not implemented" << std::endl;
    throw hMonError("Received SIGUSR1");
  }

  // handle other interrupts
  else {
    LOG(std::cerr, 0) << "ERROR: received unknown interrupt" << std::endl;
    throw hMonError("Received unknown signal");
  }

  // return to calling function
  return;

}

/******************************************************************************/
/*                                  LogStamp                                  */
/******************************************************************************/

std::string
hMon::logStamp(int verboseThreshold)
{

  // construct log stamp string
  std::ostringstream logStampStream;
  char timeStamp[24];
  logStampStream << processName
                 << ": "
                 << TimeStr(Now(), timeStamp, "%04Y-%02m-%02d %02H:%02N:%02S %Z")
                 << ": "
                 << verboseThreshold
                 << ": ";

  // return log stamp string
  return logStampStream.str();

}

/******************************************************************************/
/*                               readArguments                                */
/******************************************************************************/

void
hMon::readArguments(int argc, const char *argv[])
{

  // name of executable
  processName = argv[0];

  // initialize syntax error flag
  bool syntaxError = false;

  // begin loop over command line arguments
  for (int argumentNumber = 1; argumentNumber < argc; argumentNumber++) {

    // extract argument
    std::string argument = argv[argumentNumber];

    // handle configuration file argument
    if (argument == "-configuration") {
      configurationFile = argv[++argumentNumber];
    }

    // handle verbose level argument
    else if (argument == "-verbose") {
      verboseLevel = atoi(argv[++argumentNumber]);
    }

    // handle DatEnv arguments
    else if (isDatEnvArg(argument.c_str())) {
      argumentNumber++;
    }

    // otherwise report error
    else {
      LOG(std::cerr, 0) << "ERROR: unrecognized argument \""
                        << argument << "\"" << std::endl;
      syntaxError = true;
    }
  }

  // test that a configuration file was specified
  if (configurationFile == "") {
    LOG(std::cerr, 0) << "ERROR: configuration file not specified" << std::endl;
    syntaxError = true;
    }

  // if syntax err occured
  if (syntaxError) {

    // report error
    LOG(std::cerr, 0) << "ERROR: command syntax:" << std::endl;

    // print usage information
    std::cerr << processName << " "
              << "-configuration configurationFile "
              << "[-verbose verboseLevel] "
              << "[-partition partitionName] "
              << "[-infile frameFiles] "
              << "[-inlist listFile] "
              << "[-debug debugLevel] "
              << "[-maxframes maximumValue] "
              << std::endl;

    // throw fatal exception
    throw hMonError("Error parsing command line arguments");

  }

  // return to calling function
  return;

}

/******************************************************************************/
/*                             defaultParameters                              */
/******************************************************************************/

void
hMon::defaultParameters(void)
{

  // report status
  LOG(std::cout, 1) << "setting default parameters " << std::endl;

  // sample frequency parameters
  sampleRate = 16384;
  upSampleFactor = 4;

  // fundamental time step parameter
  unitStride = 1.0;

  // output frame file properties
  frameStride = 10;
  frameType = "<DETECTOR>_HMON";
  frameDirectory = "output/frames/<DETECTOR>";
  frameEpochDuration = 10000;
  frameChannelName = "<DETECTOR>:LSC-STRAIN";
  frameSampleRate = 16384;
  frameCompression = static_cast<FrVectRef::compression_mode>(10);
  framePrecision = "single";

  // input channel names
  sensingChannelName = "<DETECTOR>:LSC-AS_Q";
  errorChannelName = "<DETECTOR>:LSC-DARM_ERR";
  controlChannelName = "<DETECTOR>:LSC-DARM_CTRL";
  excitationChannelName = "<DETECTOR>:LSC-DARM_CTRL_EXC_DAQ";

  // input filter file names
  sensingFilterFile = "input/filters/<DETECTOR>/sensing.txt";
  servoFilterFile = "input/filters/<DETECTOR>/servo.txt";
  actuationFilterFile = "input/filters/<DETECTOR>/actuation.txt";
  lowPassFilterFile = "input/filters/<DETECTOR>/lowpass.txt";
  highPassFilterFile = "input/filters/<DETECTOR>/highpass.txt";

  // segment parameters
  segmentVersion = 1;
  segmentDirectory = "output/segments/<DETECTOR>";
  detectorStatesFile = "input/config/<DETECTOR>/states.txt";
  segmentListDuration = Interval(3.0 * 86400.0);

  // server parameters
  serverDuration = Interval(8 * 3600.0);

  // trend parameters

  // threshold for latency warning
  latencyThreshold = 0.0;

  // return to calling function
  return;

}

/******************************************************************************/
/*                             readConfiguration                              */
/******************************************************************************/

void
hMon::readConfiguration(void)
{

  // report status
  LOG(std::cout, 1) << "reading configuration file \""
                    << configurationFile << "\"" << std::endl;

  // initialize configuration file line parser
  ParseLine lineParser(configurationFile.c_str());

  // test for error opening file
  if (!lineParser.isOpen()) {
    LOG(std::cerr, 0) << "ERROR: could not open configuration file \""
                      << configurationFile << "\"" << std::endl;
    throw hMonError("Error opening configuration file");
  }

  // set field separators
  lineParser.setDelim(" \t=");

  // set comment identifiers
  lineParser.setComment("#%");

  // begin loop over over configuration file lines
  while (lineParser.getLine() >= 0) {

    // ignore blank lines
    if (lineParser.getCount() == 0) {
      continue;
    }

    // test for two words
    // ********** REMOVE THIS REQUIREMENT? **********
    if (lineParser.getCount() != 2) {
      LOG(std::cerr, 0) << "ERROR: syntax error on line "
                        << lineParser.getLineNumber() << " of configuration file \""
                        << configurationFile << "\"" << std::endl;
      for (int wordNumber = 0; wordNumber < lineParser.getCount(); wordNumber++) {
        std::cerr << "lineParser[" << wordNumber << "] =\"" << lineParser[wordNumber] << "\"" << std::endl;
      }
      throw hMonError("Syntax error in configuration file");
    }

    // extract parameter name and value
    std::string parameterName = lineParser[0];
    std::string parameterValue = lineParser[1];

    // report status
    LOG(std::cout, 3) << "read parameter \"" << parameterName
                      << "\" of \"" << parameterValue << "\"" << std::endl;

    // read detector name parameter
    if (parameterName == "detectorName") {
      detectorName = lineParser.getArg(1);
    }

    // read sample frequency parameters
    else if (parameterName == "sampleRate") {
      sampleRate = lineParser.getInt(1);
    }
    else if (parameterName == "upSampleFactor") {
      upSampleFactor = lineParser.getInt(1);
    }

    // read fundamental time step parameter
    else if (parameterName == "unitStride") {
      unitStride = lineParser.getDouble(1);
    }

    // read output frame parameters
    else if (parameterName == "frameStride") {
      frameStride = lineParser.getDouble(1);
    }
    else if (parameterName == "frameType") {
      frameType = lineParser.getArg(1);
    }
    else if (parameterName == "frameDirectory") {
      frameDirectory = lineParser.getArg(1);
    }
    else if (parameterName == "frameEpochDuration") {
      frameEpochDuration = lineParser.getInt(1);
    }
    else if (parameterName == "frameChannelName") {
      frameChannelName = lineParser.getArg(1);
    }
    else if (parameterName == "frameSampleRate") {
      frameSampleRate = lineParser.getInt(1);
    }
    else if (parameterName == "framePrecision") {
      framePrecision = lineParser.getArg(1);
    }
    else if (parameterName == "frameCompression") {
      frameCompression = static_cast<FrVectRef::compression_mode>(lineParser.getInt(1));
    }

    // read detector state parameters
    else if (parameterName == "detectorStatesFile") {
      detectorStatesFile = lineParser.getArg(1);
    }

    // read segment parameters
    else if (parameterName == "segmentVersion") {
      segmentVersion = lineParser.getInt(1);
    }
    else if (parameterName == "segmentDirectory") {
      segmentDirectory = lineParser.getArg(1);
    }
    else if (parameterName == "segmentListDuration") {
      segmentListDuration = lineParser.getDouble(1);
    }

    // read input channel name parameters
    else if (parameterName == "sensingChannelName") {
      sensingChannelName = lineParser.getArg(1);
    }
    else if (parameterName == "errorChannelName") {
      errorChannelName = lineParser.getArg(1);
    }
    else if (parameterName == "controlChannelName") {
      controlChannelName = lineParser.getArg(1);
    }
    else if (parameterName == "excitationChannelName") {
      excitationChannelName = lineParser.getArg(1);
    }

    // read filter file name parameters
    else if (parameterName == "sensingFilterFile") {
      sensingFilterFile = lineParser.getArg(1);
    }
    else if (parameterName == "servoFilterFile") {
      servoFilterFile = lineParser.getArg(1);
    }
    else if (parameterName == "actuationFilterFile") {
      actuationFilterFile = lineParser.getArg(1);
    }
    else if (parameterName == "lowPassFilterFile") {
      lowPassFilterFile = lineParser.getArg(1);
    }
    else if (parameterName == "highPassFilterFile") {
      highPassFilterFile = lineParser.getArg(1);
    }

    // read excitation line parameters
    else if (parameterName == "excitationFrequency") {
      excitationFrequencies.push_back(lineParser.getDouble(1));
    }
    else if (parameterName == "excitationAmplitude") {
      excitationAmplitudes.push_back(lineParser.getDouble(1));
    }

    // read latency warning parameters
    else if (parameterName == "latencyThreshold") {
      latencyThreshold = lineParser.getDouble(1);
    }

    // read spectrum parameters
    else if (parameterName == "spectrumResolution") {
      spectrumResolution = lineParser.getDouble(1);
    }
    else if (parameterName == "spectrumDecayTime") {
      spectrumDecayTime = lineParser.getDouble(1);
    }

    // read server parameters
    else if (parameterName == "serverDuration") {
      serverDuration = lineParser.getDouble(1);
    }

    // trend parameters

    // handle unknown parameters
    else {
      LOG(std::cerr, 0) << "ERROR: unknown parameter "
                        << parameterName << " on line "
                        << lineParser.getLineNumber() << " of configuration file \""
                        << configurationFile << "\"" << std::endl;
      throw hMonError("Error reading configuration file");
    }

  // end loop over configuration file lines
  }

  // return to calling function
  return;

}

/******************************************************************************/
/*                             derivedParameters                              */
/******************************************************************************/

// utility function for template replacement in parameters
void
replace(std::string& parameter, std::string identifier, std::string value)
{
  while (parameter.find(identifier) != std::string::npos) {
    parameter.replace(parameter.find(identifier), identifier.size(), value);
  }
}

void
hMon::derivedParameters(void)
{

  // report status
  LOG(std::cout, 1) << "determining derived parameters " << std::endl;

  // detector specific process name
  monitorName = processName + "_" + detectorName;

  // extract site name from detector name
  siteName = detectorName[0];

  // weighting factor for averaging spectra
  spectrumWeightingFactor = 1 - exp(unitStride / spectrumDecayTime);

  // perform detector Name substitution
  replace(frameType, "<DETECTOR>", detectorName);
  replace(frameChannelName, "<DETECTOR>", detectorName);
  replace(sensingChannelName, "<DETECTOR>", detectorName);
  replace(errorChannelName, "<DETECTOR>", detectorName);
  replace(controlChannelName, "<DETECTOR>", detectorName);
  replace(excitationChannelName, "<DETECTOR>", detectorName);
  replace(sensingFilterFile, "<DETECTOR>", detectorName);
  replace(servoFilterFile, "<DETECTOR>", detectorName);
  replace(actuationFilterFile, "<DETECTOR>", detectorName);
  replace(lowPassFilterFile, "<DETECTOR>", detectorName);
  replace(highPassFilterFile, "<DETECTOR>", detectorName);
  replace(frameDirectory, "<DETECTOR>", detectorName);
  replace(segmentDirectory, "<DETECTOR>", detectorName);
  replace(detectorStatesFile, "<DETECTOR>", detectorName);

  // return to calling function
  return;

}

/******************************************************************************/
/*                             validateParameters                             */
/******************************************************************************/

void
hMon::validateParameters(void)
{

  // report status
  LOG(std::cout, 1) << "validating parameters " << std::endl;

  // test for valid detector name
  if (detectorName != "H1" &&
      detectorName != "H2" &&
      detectorName != "L1") {
    LOG(std::cerr, 0) << "ERROR: unknown detector \""
                      << detectorName << "\""
                      << std::endl;
    throw hMonError("Invalid parameter");
  }

  // test for integer second unit stride
  if (fmod(unitStride, 1.0) != 0.0) {
    LOG(std::cerr, 0) << "ERROR: unit stride of " << unitStride
                      << " seconds is not an integer" << std::endl;
    throw hMonError("Invalid parameter");
  }

  // test for integer second frame stride
  if (fmod(frameStride, 1.0) != 0.0) {
    LOG(std::cerr, 0) << "ERROR: frame stride of " << frameStride
                      << " seconds is not an integer" << std::endl;
    throw hMonError("Invalid parameter");
  }

  // test for writable segment directory
  if (access(segmentDirectory.c_str(), W_OK) != 0) {
    if (mkdir(segmentDirectory.c_str(), 0755) != 0) {
      LOG(std::cerr, 0) << "ERROR: invalid segment directory \""
                        << segmentDirectory << "\"" << std::endl;
      throw hMonError("Invalid parameter");
    }
  }

  // test for writeable frame directory
  if (access(frameDirectory.c_str(), W_OK) != 0) {
    if (mkdir(frameDirectory.c_str(), 0755) != 0) {
      LOG(std::cerr, 0) << "ERROR: invalid frame directory \""
                        << frameDirectory << "\"" << std::endl;
      throw hMonError("Invalid parameter");
    }
  }

  // return to calling function
  return;

}

/******************************************************************************/
/*                               initializeData                               */
/******************************************************************************/

void
hMon::initializeData(void)
{

  // report status
  LOG(std::cout, 1) << "initializing data accessor and containers" << std::endl;

  // set length of data containers
  sensingChannelData.setMaxLen(frameStride);
  errorChannelData.setMaxLen(frameStride);
  controlChannelData.setMaxLen(frameStride);
  excitationChannelData.setMaxLen(frameStride);

  // set length of optical gain factor time series
  alphaTimeSeries.setMaxLen(frameStride);

  // initialize the data accessor
  getDacc().setStride(unitStride);
  getDacc().addChannel(sensingChannelName.c_str());
  getDacc().addChannel(errorChannelName.c_str());
  getDacc().addChannel(controlChannelName.c_str());
  getDacc().addChannel(excitationChannelName.c_str());

  // return to calling function
  return;

}

/******************************************************************************/
/*                             initializeSpectrum                             */
/******************************************************************************/

void
hMon::initializeSpectrum(void)
{

  // report status
  LOG(std::cout, 1) << "initializing calibrated spectrum containers" << std::endl;

  // initialize average spectrum
  averageSpectrum.clear();

  // initialize window
  // ********** COMPUTE SPECTRA AT FRAMESAMPLERATE? **********
  spectrumWindow.setWindow(frameStride.GetS() * sampleRate);

  // return to calling function
  return;

}

/******************************************************************************/
/*                              initializeServer                              */
/******************************************************************************/

void
hMon::initializeServer(void)
{

  // report status
  LOG(std::cout, 1) << "initializing dmtviewer server" << std::endl;

  // set monitor name as seen in dmt viewer
  setServerName(monitorName.c_str());

  // set time range of history data to serve
  alphaHistory.setMaxLen(serverDuration);
  horizonDistanceHistory.setMaxLen(serverDuration);
  senseMonRangeHistory.setMaxLen(serverDuration);
  figureOfMeritHistory.setMaxLen(serverDuration);

  // setup alpha history for serving
  serveData("optical_gain_factor", &alphaHistory, 0);

  // setup horizon distance for servier
  serveData("horizon_distance", &horizonDistanceHistory, 0);

  // setup sensemon range for serving
  serveData("sensemon_range", &senseMonRangeHistory, 0);

  // setup ligo design figure of merit for serving
  serveData("design_figure_of_merit", &figureOfMeritHistory, 0);

  // setup calibrated spectrum for serving
  serveData("calibrated_spectrum", &averageSpectrum, 0);

  // return to calling function
  return;

}

/******************************************************************************/
/*                              initializeTrends                              */
/******************************************************************************/

void
hMon::initializeTrends(void)
{

  // report status
  LOG(std::cout, 1) << "initializing minute and second trenders" << std::endl;

  // initialize alpha minute trender
  alphaMinuteTrend.setName(monitorName.c_str());
  alphaMinuteTrend.setFrameCount(1);
  alphaMinuteTrend.setType(Trend::kMinute);
  alphaMinuteTrend.setIFO(detectorName.c_str());
  alphaMinuteTrend.setAutoUpdate(false);
  std::string trendChannelName = detectorName + ":DMT-HMON_LSC_STRAIN_alpha";
  alphaMinuteTrend.addChannel(trendChannelName.c_str());
  alphaMinuteTrend.writeIndex();

  // return to calling function
  return;

}

/******************************************************************************/
/*			       initializeFilters                              */
/******************************************************************************/

void
hMon::initializeFilters(void)
{

  // report status
  LOG(std::cout, 1) << "reading filters" << std::endl;

  // read inverse sensing filter
  LOG(std::cout, 2) << "reading inverse sensing filter" << std::endl;
  readFilterFile(sensingFilterFile, sensingFilter,
                 sampleRate * upSampleFactor);

  // read inverse servo filter
  LOG(std::cout, 2) << "reading inverse servo filter" << std::endl;
  readFilterFile(servoFilterFile, servoFilter,
                 sampleRate * upSampleFactor);

  // read inverse actuation filter
  LOG(std::cout, 2) << "reading inverse actuation filter" << std::endl;
  readFilterFile(actuationFilterFile, actuationFilter,
                 sampleRate * upSampleFactor);

  // read low pass filter
  LOG(std::cout, 2) << "reading low pass filter" << std::endl;
  readFilterFile(lowPassFilterFile, lowPassFilter, sampleRate);

  // read high pass filter
  LOG(std::cout, 2) << "reading high pass filter" << std::endl;
  readFilterFile(highPassFilterFile, highPassFilter, sampleRate);

  // determine maximum filter length
  LOG(std::cout, 2) << "determining maximum filter delay" << std::endl;
  filterDelay = Interval(0.0);
  if (sensingFilter.getLength() > filterDelay) {
    filterDelay = sensingFilter.getTimeDelay();
  }
  if (servoFilter.getLength() > filterDelay) {
    filterDelay = servoFilter.getTimeDelay();
  }
  if (actuationFilter.getLength() > filterDelay) {
    filterDelay = actuationFilter.getTimeDelay();
  }
  if (lowPassFilter.getLength() > filterDelay) {
    filterDelay = lowPassFilter.getTimeDelay();
  }
  if (highPassFilter.getLength() > filterDelay) {
    filterDelay = highPassFilter.getTimeDelay();
  }

  // round up to nearest integer second
  LOG(std::cout, 2) << "rounding maximum filter delay up to nearest second" << std::endl;
  if (filterDelay.GetN() != 0) {
    filterDelay = static_cast<Interval>(filterDelay.GetS() + 1);
  }

  // report effective filter delay
  LOG(std::cout, 2) << "filter delay: " << filterDelay << std::endl;

  // test for short stride
  if (frameStride < 2.0 * filterDelay) {
    LOG(std::cout, 0) << ": WARNING: " << frameStride.GetS()
                      << " second frame stride is smaller than "
                      << 2 * filterDelay.GetS()
                      << " second filter duration." << std::endl;
  }

  // return to calling function
  return;

}

/******************************************************************************/
/*                               readFilterFile                               */
/******************************************************************************/

void
hMon::readFilterFile(const std::string& filterFile, FIRdft& filter, int filterRate)
{

  // report status
  LOG(std::cout, 3) << "reading filter file \"" << filterFile << "\"" << std::endl;

  // initialize coefficient vector
  DVecType<double> coefficients;

  // initialize filter file line parser
  ParseLine lineParser(filterFile.c_str());

  // test for error opening file
  if (!lineParser.isOpen()) {
    LOG(std::cerr, 0) << "ERROR: could not open filter file \""
                     << filterFile << "\"" << std::endl;
    throw hMonError("Error opening filter file");
  }

  // set field separators
  lineParser.setDelim(" \t=:");

  // set comment identifiers
  lineParser.setComment("#%/");

  // begin loop over over configuration file lines
  while (lineParser.getLine() >= 0) {

    // ignore blank lines
    if (lineParser.getCount() == 0) {
      continue;
    }

    // test for more than one word
    if (lineParser.getCount() != 1) {
      LOG(std::cerr, 0) << "ERROR: syntax error on line "
                        << lineParser.getLineNumber() << " of filter file \""
                        << filterFile << "\"" << std::endl;
      throw hMonError("Error reading filter file");
    }

    // read coefficient from file
    double coefficient = lineParser.getDouble(0);

    // insert coefficient into coefficient vector
    coefficients.Append(1, &coefficient);

  // end loop over filter file lines
  }

  // report filter properties
  LOG(std::cout, 3) << "filter order: " << coefficients.getLength() - 1 << std::endl;
  LOG(std::cout, 3) << "filter rate: " << filterRate << std::endl;
  LOG(std::cout, 4) << "filter coefficients:" << std::endl;
  if (verboseLevel >= 4) {
    coefficients.Dump(std::cout);
  }

  // round duration up to nearest integer second
  // ********** IS THIS NECESSARY? **********
  LOG(std::cout, 2) << "zero padding filter to nearest integer second" << std::endl;
  int filterOrder = coefficients.getLength() - 1;
  int filterDuration = filterOrder / filterRate;
  if (filterOrder % filterRate != 0) {
    filterDuration += 1;
  }
  filterOrder = filterDuration * filterRate;
  int filterLength = filterOrder + 1;

  // zero pad coefficients to new filter length
  coefficients.Extend(filterLength);

  // set filter length and coefficients
  filter.setCoefs(coefficients.getLength(), coefficients.refTData());

  // set filter to zero phase mode
  filter.setMode(FIRdft::fm_zero_phase);

  // set sample frequency of filter
  filter.setRate(filterRate);

  // report filter properties
  LOG(std::cout, 3) << "filter order: " << filter.getLength() - 1 << std::endl;
  LOG(std::cout, 3) << "filter rate: " << filter.getRate() << std::endl;
  LOG(std::cout, 3) << "filter delay: " << filter.getTimeDelay().GetS()
                    << " seconds" << std::endl;
  LOG(std::cout, 4) << "filter coefficients:" << std::endl;
  if (verboseLevel >= 4) {
    coefficients.Dump(std::cout);
  }

  // return to calling function
  return;

}

/******************************************************************************/
/*				initializeStates                              */
/******************************************************************************/

void
hMon::initializeStates(void)
{

  // report status
  LOG(std::cout, 1) << "reading detector state definitions" << std::endl;

  // set stride
  detectorStates.setStride(unitStride);

  // set debug level
  detectorStates.setDebug(0);

  // test that detector states definition file exists
  if (access(detectorStatesFile.c_str(), R_OK) != 0) {
    LOG(std::cerr, 0) << "ERROR: invalid detector states definition file \""
                      << detectorStatesFile << "\"" << std::endl;
    throw hMonError("Invalid detector states definition file");
  }

  // read detector state definition file
  detectorStates.readConfig(detectorStatesFile);

  // test for definition of lock state
  if (detectorStates.find("LOCK") == detectorStates.end()) {
    LOG(std::cerr, 0) << "ERROR: detector state LOCK not defined in \""
                      << detectorStatesFile << "\"" << std::endl;
    throw hMonError("Error reading detector state definitions file");
  }

  // test for definition of science state
  if (detectorStates.find("SCIENCE") == detectorStates.end()) {
    LOG(std::cerr, 0) << "ERROR: detector state SCIENCE not defined in \""
                      << detectorStatesFile << "\"" << std::endl;
    throw hMonError("Error reading detector state definitions file");
  }

  // test for definition of injection state
  if (detectorStates.find("INJECTION") == detectorStates.end()) {
    LOG(std::cerr, 0) << "ERROR: detector state INJECTION not defined in \""
                      << detectorStatesFile << "\"" << std::endl;
    throw hMonError("Error reading detector state definitions file");
  }

  // test for definition of saturation state
  if (detectorStates.find("SATURATION") == detectorStates.end()) {
    LOG(std::cerr, 0) << "ERROR: detector state SATURATION not defined in \""
                      << detectorStatesFile << "\"" << std::endl;
    throw hMonError("Error reading detector state definitions file");
  }

  // return to calling function
  return;

}

/******************************************************************************/
/*                             initializeSegments                             */
/******************************************************************************/

void
hMon::initializeSegments(void)
{

  // report status
  LOG(std::cout, 1) << "initializing segment containers" << std::endl;

  // initialize last read time
  lastReadTime = Time(0);

  // create segment lists
  segmentLists["read"];
  segmentLists["missing"];
  segmentLists["transient"];
  segmentLists["bad"];
  segmentLists["lock"];
  segmentLists["science"];
  segmentLists["injection"];
  segmentLists["saturation"];
  segmentLists["calibrated"];
  segmentLists["veto"];

  // initialize segment lists
  ManagedSegmentListMapIterator segmentListsIterator;
  for (segmentListsIterator = segmentLists.begin();
       segmentListsIterator != segmentLists.end();
       segmentListsIterator++) {
    std::string listName = segmentListsIterator->first;
    ManagedSegmentList* segmentList = &segmentListsIterator->second;
    segmentList->setMaximumDuration(segmentListDuration);
    segmentList->setRecentDuration(frameStride + filterDelay);
    // segmentList->setRecentDuration(calibrationStride + filterDelay); **********
    segmentList->setFileName(segmentDirectory + "/" + listName + ".txt");
    segmentList->setListName(detectorName + "_" + listName);
    segmentList->setListDescription(detectorName + "_" + listName);
    segmentList->addDispatchMode("txt");
    segmentList->addDispatchMode("sql");
    segmentList->reset();
  }

  // return to calling function
  return;

}

/******************************************************************************/
/*                                  readData                                  */
/******************************************************************************/

void
hMon::readData(void)
{

  // get start time of new data
  currentReadTime = getDacc().getFillTime();

  // report status
  LOG(std::cout, 3) << "reading " << detectorName
                    << " data at " << currentReadTime.getS() << std::endl;

  // read sensing channel data
  readChannel(sensingChannelName, sensingChannelData);

  // read error channel data
  readChannel(errorChannelName, errorChannelData);

  // read control channel data
  readChannel(controlChannelName, controlChannelData);

  // read actuation channel data
  readChannel(excitationChannelName, excitationChannelData);

  // return to calling function
  return;

}

/******************************************************************************/
/*                                readChannel                                 */
/******************************************************************************/

void
hMon::readChannel(const std::string& channelName, FixedLenTS& channelData)
{

  // report status
  LOG(std::cout, 3) << "reading " << channelName << "data" << std::endl;

  // get pointer to new data
  const TSeries* timeSeries = getDacc().refData(channelName.c_str());
  if (!timeSeries || timeSeries->isEmpty()) {
    LOG(std::cerr, 0) << "ERROR: could not read channel \"" << channelName
                      << "\"" << std::endl;
    throw hMonWarning("Error reading data");
    return;
  }

  // test for consistent sample frequency
  if (timeSeries->getNSample() !=
      static_cast<size_t>(unitStride.GetS() * sampleRate)) {
    LOG(std::cerr, 0) << "ERROR: inconsistent data length when reading channel "
                      << "\"" << channelName << "\"" << std::endl;
    throw hMonWarning("Error reading data");
    return;
  }

  // fill missing data with zeros
  channelData.padUntil(currentReadTime, timeSeries->getTStep());

  // append data to fixed length tseries
  channelData.fixedAppend(*timeSeries);

  // return to calling function
  return;

}

/******************************************************************************/
/*                                testLatency                                 */
/******************************************************************************/

void
hMon::testLatency(void)
{

  // get real time
  realTime = Now();

  // if non-zero latency threshold requested
  if (latencyThreshold > Interval(0.0)) {

    // report status
    LOG(std::cout, 2) <<  "testing latency" << std::endl;

    // test for excessive latency
    if (currentReadTime < realTime - latencyThreshold) {

      // ********** ISSUE LATENCY ALARM **********

      // report warning
      LOG(std::cerr, 1) << "WARNING: current read time "
                        << currentReadTime.getS() << " more than "
                        << latencyThreshold.GetS() << " seconds before real time "
                        << realTime.getS() << std::endl;

    // end test for excessive latency
    }

  // end test for non-zero latency threhsold
  }

  // return to calling function
  return;

}

/******************************************************************************/
/*                               updateSegments                               */
/******************************************************************************/

void
hMon::updateSegments(void)
{

  // test for gaps in data
  if (lastReadTime != currentReadTime) {

    // reset filter history
    resetFilters();

    // reset spectrum computation
    resetSpectrum();

    // record current start transient
    segmentLists["transient"].insert(currentReadTime, filterDelay);    

    // test for gap in processing
    if (lastReadTime != Time(0)) {

      // report missing data
      LOG(std::cout, 1) << "WARNING: missing data between "
                        << lastReadTime.getS()
                        << " and "
                        << currentReadTime.getS()
                        << std::endl;

      // record missing segment
      segmentLists["missing"].insert(lastReadTime, currentReadTime);

      // record previous stop transient
      segmentLists["transient"].insert(lastReadTime - filterDelay, filterDelay);

    }

    // handle start of processing
    else {

      // record missing segment
      // (we don't want to write this to a file or publish it)
      // segmentLists["missing"].insert(Time(0), currentReadTime);

      // report processing start time
      LOG(std::cout, 1) << "processing starting with "
                        << currentReadTime.getS()
                        << std::endl;

    // end test for gaps in processing
    }

  }

  // build read segment list
  segmentLists["read"].insert(currentReadTime, unitStride);

  // build locked segment list
  if (detectorStates.satisfied("LOCK")) {
    LOG(std::cout, 4) << detectorName << " in lock at "
                      << currentReadTime.getS() << std::endl;
    segmentLists["lock"].insert(currentReadTime, unitStride);
  }

  // build science mode segment list
  if (detectorStates.satisfied("SCIENCE")) {
    LOG(std::cout, 4) << detectorName << " in science mode at "
                      << currentReadTime.getS() << std::endl;
    segmentLists["science"].insert(currentReadTime, unitStride);
  }

  // build injection mode segment list
  if (detectorStates.satisfied("INJECTION")) {
    LOG(std::cout, 4) << detectorName << " in injection mode at "
                      << currentReadTime.getS() << std::endl;
    segmentLists["injection"].insert(currentReadTime, unitStride);
  }

  // build injection mode segment list
  if (detectorStates.satisfied("SATURATION")) {
    LOG(std::cout, 4) << detectorName << " in saturation at "
                      << currentReadTime.getS() << std::endl;
    segmentLists["saturation"].insert(currentReadTime, unitStride);
  }

  // update last read time
  lastReadTime = currentReadTime + unitStride;

  // return to calling function
  return;

}

/******************************************************************************/
/*                              finalizeSegments                              */
/******************************************************************************/

void
hMon::finalizeSegments(void)
{

  // record transient segment
  segmentLists["transient"].insert(lastReadTime - filterDelay, filterDelay);

  // record missing segment
  // (we don't want to write this to a file or publish it)
  // segmentLists["missing"].insert(lastReadTime, Time(SEGMENT_TINF));

  // return to calling function
  return;

}

/******************************************************************************/
/*                                resetFilters                                */
/******************************************************************************/

void
hMon::resetFilters(void)
{

  // report status
  LOG(std::cout, 2) <<  "resetting filters histories" << std::endl;

  // reset filter history
  sensingFilter.reset();
  servoFilter.reset();
  actuationFilter.reset();
  lowPassFilter.reset();
  highPassFilter.reset();

  // return to calling function
  return;

}

/******************************************************************************/
/*                               resetSpectrum                                */
/******************************************************************************/

void
hMon::resetSpectrum(void)
{

  // report status
  LOG(std::cout, 2) << "resetting calibrated spectrum containers" << std::endl;

  // reset average spectrum
  averageSpectrum.clear();

  // reset window
  spectrumWindow.reset();

  // return to calling function
  return;

}

/******************************************************************************/
/*                               isFrameStride                                */
/******************************************************************************/

bool
hMon::isFrameStride(void)
{

  // adjusted start time of data to calibrate
  calibratedStartTime = lastReadTime - frameStride - filterDelay;

  // compute offset from integer multiple of frame stride
  long remainder = static_cast<long>(calibratedStartTime.getS()) %
    frameStride.GetS();

  // return true if stride start time is aligned with frame stride
  if (remainder == 0) {
    return true;
  }
  else {
    return false;
  }

}

/******************************************************************************/
/*                               computeFactors                               */
/******************************************************************************/

void
hMon::computeFactors(void)
{

  // compute optical gain factor
  alpha = 1.2345;

  // construct optical gain factor time series
  alphaTimeSeries.fixedAppend(currentReadTime, unitStride, &alpha);

  // return to calling function
  return;

}

/******************************************************************************/
/*                               calibrateData                                */
/******************************************************************************/

void
hMon::calibrateData(void)
{

  // filter data
  sensingFilter.apply(sensingChannelData, calibratedData);

  // apply factors

  // sum

  // apply high pass filter
  // ********** need to add delay to filterDelay? **********
  // highPassFilter.apply(calibratedData, calibratedData);

  // apply low pass filter
  // ********** need to add delay to filterDelay? **********
  // lowPassFilter.apply(calibratedData, calibratedData);

  // down sample

  // stop time of most recently calibrated data
  lastCalibratedTime = lastReadTime - filterDelay;

  // create segment lists
  segmentLists["read"].update(lastCalibratedTime);
  segmentLists["missing"].update(lastCalibratedTime);
  segmentLists["transient"].update(lastCalibratedTime);
  segmentLists["bad"].update(lastCalibratedTime);
  segmentLists["lock"].update(lastCalibratedTime);
  segmentLists["science"].update(lastCalibratedTime);
  segmentLists["injection"].update(lastCalibratedTime);
  segmentLists["saturation"].update(lastCalibratedTime);
  segmentLists["calibrated"].update(lastCalibratedTime);
  segmentLists["veto"].update(lastCalibratedTime);

  // udpate calibrated segment list
  LockSegList calibratedSegments;
  calibratedSegments |= segmentLists["lock"].getRecentSegments();
  calibratedSegments.pad(-filterDelay, -filterDelay);
  LockSegment maskSegment(0, lastCalibratedTime - frameStride, frameStride);
  // LockSegment maskSegment(0, lastCalibratedTime - calibrationStride, calibrationStride);
  LockSegList maskSegments;
  maskSegments.insert(maskSegment);
  calibratedSegments &= maskSegments;
  segmentLists["calibrated"].insert(calibratedSegments);

  // update veto segment list
  LockSegList vetoSegments;
  // ********** avoid bug in LockSegment::pad() **********
  // vetoSegments |= segmentLists["science"].getRecentSegments();
  // vetoSegments.invert();
  vetoSegments |= segmentLists["bad"].getRecentSegments();
  vetoSegments |= segmentLists["injection"].getRecentSegments();
  vetoSegments |= segmentLists["saturation"].getRecentSegments();
  vetoSegments.pad(+filterDelay, +filterDelay);
  LockSegList nonScienceSegments;
  nonScienceSegments |= segmentLists["science"].getRecentSegments();
  nonScienceSegments.invert();
  vetoSegments |= nonScienceSegments;
  vetoSegments &= calibratedSegments;
  segmentLists["veto"].insert(vetoSegments);

  // return to calling function
  return;

}

/******************************************************************************/
/*                              computeSpectrum                               */
/******************************************************************************/

void
hMon::computeSpectrum(void)
{

  // report status
  LOG(std::cout, 3) << "computing spectrum" << std::endl;

  // apply time domain window
  currentSpectrum.setData(spectrumWindow.apply(calibratedData));

  // return to calling function
  return;

}

/******************************************************************************/
/*                             computeSensitivity                             */
/******************************************************************************/

void
hMon::computeSensitivity(void)
{

  // report status
  LOG(std::cout, 3) << "computing sensitivity" << std::endl;

  // compute horizon distance to fiducial source
  horizonDistance = 1.2345;

  // compute sensemon distance to fiducial source
  senseMonRange = 1.2345;

  // compute ligo design figure of merit
  // ********** BAND SELETED BASED ON INITIAL DESIGN NOISE CURVE **********
  // ********** DYNAMICALLY CHOOSE 100 HZ BAND? **********
  figureOfMerit = sqrt(averageSpectrum.getSum(124.0, 224.0) *
                       averageSpectrum.getFStep());

  // report rms signal in initial ligo band
  LOG(std::cout, 4) << "ligo design figure of merit: "
                    << figureOfMerit
                    << " strain" << std::endl;

  // return to calling function
  return;

}

/******************************************************************************/
/*                                updateServer                                */
/******************************************************************************/

void
hMon::updateServer(void)
{

  // update optical gain factor
  alphaHistory.fixedAppend(currentReadTime, frameStride, &alpha);

  // update horizon distance
  horizonDistanceHistory.fixedAppend(currentReadTime, frameStride, &horizonDistance);

  // udpate sensemon range
  senseMonRangeHistory.fixedAppend(currentReadTime, frameStride, &senseMonRange);

  // update ligo design figure of merit
  figureOfMeritHistory.fixedAppend(currentReadTime, frameStride, &figureOfMerit);

  // update average spectrum
  averageSpectrum *= (1 / spectrumWeightingFactor - 1);
  averageSpectrum += currentSpectrum;
  averageSpectrum *= spectrumWeightingFactor;

  // return to calling function
  return;

}

/******************************************************************************/
/*                               publishFrames                                */
/******************************************************************************/

void
hMon::publishFrames()
{

  // name of output directory
  std::ostringstream outputDirectoryName;
  long epoch = calibratedStartTime.getS() / frameEpochDuration;
  outputDirectoryName << frameDirectory << "/"
                      << siteName << "-"
                      << frameType << "-"
                      << epoch;

  // name of frame file to write to
  std::ostringstream frameFileName;
  frameFileName << outputDirectoryName.str() << "/"
                << siteName << "-"
                << frameType << "-"
                << calibratedStartTime.getS() << "-"
                << frameStride.GetS() << ".gwf";

  // report status
  LOG(std::cout, 1) << "writing " << frameFileName.str() << std::endl;

  // cast to single precision if requested
  if (framePrecision == "single") {
    calibratedData.Convert(DVector::t_float);
  }
  else if (framePrecision != "double") {
    LOG(std::cerr, 0) << "ERROR: unknown frame precision \""
                      << framePrecision << "\"" << std::endl;
    throw hMonError("Unknown precision type");
  }

  // create output directory if necessary
  if (access(outputDirectoryName.str().c_str(), W_OK) != 0) {
    if (mkdir(outputDirectoryName.str().c_str(), 0755) != 0) {
      LOG(std::cerr, 0) << "ERROR: unable to create directory \""
                        << outputDirectoryName.str() << "\"" << std::endl;
      throw hMonWarning("Unable to create output directory");
    }
  }

  // write calibrate time series frame data
  frameWriter.buildFrame(calibratedData.getStartTime(), frameStride);
  frameWriter.addProcSeries(frameChannelName.c_str(), calibratedData);
  frameWriter.addHistory(monitorName, Now(), "real-time time-domain calibrated data");
  frameWriter.addWriterHistory();
  frameWriter.setCompress(frameCompression);
  frameWriter.open(frameFileName.str().c_str(), true);
  frameWriter.writeFrame();
  frameWriter.close();

  // return to calling function
  return;

}

/******************************************************************************/
/*                              publishSegments                               */
/******************************************************************************/

void
hMon::publishSegments(void)
{

  // send new segments to database
  sendSegments("read");  			  // ********** TEMPORARY
  sendSegments("missing");
  sendSegments("transient");  			  // ********** TEMPORARY
  sendSegments("bad");
  sendSegments("lock");
  sendSegments("science");
  sendSegments("injection");
  sendSegments("saturation");
  sendSegments("calibrated");
  sendSegments("veto");

  // write segment lists to disk
  writeSegments("read");  			  // ********** TEMPORARY
  writeSegments("missing");  			  // ********** TEMPORARY
  writeSegments("transient");  			  // ********** TEMPORARY
  writeSegments("bad");
  writeSegments("lock");
  writeSegments("science");
  writeSegments("injection");
  writeSegments("saturation");
  writeSegments("calibrated");
  writeSegments("veto");

  // return to calling function
  return;

}

/******************************************************************************/
/*				 writeSegments                                */
/******************************************************************************/

void
hMon::writeSegments(std::string segmentListName)
{

  // determine segment file name
  std::string segmentFileName = segmentDirectory + "/" + segmentListName + ".txt";

  // report status
  LOG(std::cout, 2) << "writing " << segmentListName << " segments list to "
                    << segmentFileName << std::endl;

  // get reference to segment list to write
  LockSegList segmentList =
    segmentLists[segmentListName].getWriteSegments();

  // open segment list file for writing
  std::ofstream segmentFileStream(segmentFileName.c_str());

  // test for error opening segment list file
  if (!segmentFileStream.is_open()) {
    LOG(std::cerr, 0) << "ERROR: cannot open segment file \""
                      << segmentFileName << "\"" << std::endl;
    return;
  }

  // set output format
  segmentFileStream << std::right;
  segmentFileStream << std::fixed;
  segmentFileStream << std::setfill('0');

  // write header
  segmentFileStream << "# hMon segment list"
		    << std::endl;
  segmentFileStream << "# detector name: "
		    << detectorName << std::endl;
  segmentFileStream << "# segment list: "
		    << segmentListName << std::endl;
  segmentFileStream << "# segment file: "
		    << segmentFileName << std::endl;
  segmentFileStream << "# number segments: "
		    << segmentList.size() << std::endl;
  segmentFileStream << "# live time: "
		    << segmentList.live() << std::endl;

  // begin loop over segments
  for (size_t segmentNumber = 0;
       segmentNumber < segmentList.size();
       segmentNumber++) {

    // write segment start time
    segmentFileStream << std::setw(10)
		      << segmentList[segmentNumber].getStartTime().getS()
		      << "."
		      << std::setw(9)
		      << segmentList[segmentNumber].getStartTime().getN();

    // write delimiter
    segmentFileStream << " ";

    // write segment stop time
    segmentFileStream << std::setw(10)
		      << segmentList[segmentNumber].getEndTime().getS()
		      << "."
		      << std::setw(9)
		      << segmentList[segmentNumber].getEndTime().getN();

    // write end of line
    segmentFileStream << std::endl;

  // end loop over segments
  }

  // ********** PRINT LAST VALID TIME OF SEGMENT LIST **********

  // write footer
  segmentFileStream << "# last read time: "
                    << lastReadTime.getS() << std::endl;
  segmentFileStream << "# last update time: "
                    << realTime.getS() << std::endl;

  // close segment list file
  segmentFileStream.close();

  // return to calling function
  return;

}

/******************************************************************************/
/*                                sendSegments                                */
/******************************************************************************/

void
hMon::sendSegments(std::string segmentListName)
{

  // report status
  LOG(std::cout, 2) << "sending " << segmentListName << " segments"
		    << std::endl;

  // define group name based on process name and list name
  std::string groupName = processName + "_" + segmentListName;

  // convert group name to upper case
  std::transform(groupName.begin(), groupName.end(), groupName.begin(),
		 (int(*)(int)) std::toupper);

  // define comment based on detector name and group name
  std::string comment = detectorName + "_" + groupName;

  // get reference to segment list to write
  LockSegList segmentList =
    segmentLists[segmentListName].getUnsentSegments();

  // begin loop over segments
  for (size_t segmentNumber = 0;
       segmentNumber < segmentList.size();
       segmentNumber++) {

    // build segment for sending
    trig::Segment
      segment(groupName, segmentVersion,
	      segmentList[segmentNumber].getStartTime(),
	      segmentList[segmentNumber].getEndTime());
    segment.setIfos(detectorName.c_str());
    segment.setComment(comment);

    // send segment to database
    lmsg::error_type errorCode = sendSegment(segment);

    // handle errors
    if (errorCode) {
      LOG(std::cout, 0) << "error sending segment to database" << std::endl;
      throw hMonWarning("Error sending segment to database");
    }

  // end loop over segments
  }

  // mark segments as sent
  segmentLists[segmentListName].markSent();

  // return to calling function
  return;

}

/******************************************************************************/
/*                                updateTrends                                */
/******************************************************************************/

void
hMon::updateTrends(void)
{

  // trend optical gain factor
  std::string trendChannelName = detectorName + ":DMT-HMON_LSC_STRAIN_alpha";
  alphaMinuteTrend.trendData(trendChannelName.c_str(), currentReadTime, alpha);
  alphaMinuteTrend.Update();

  // return to calling function
  return;

}
